from math import radians, cos, sin, asin, sqrt
import json


# FIXME: encapsulate any concern about topology here


######################################################################################################################
# General functions
######################################################################################################################
def create_topology():
    return {
        'version': '1.0',
        'sensors': {},
        'fog': {},
        'cloud': {},
        'delay_sf': {},
        'delay_fc': {},
        'delay_ff': {},
        'conn_sf': {},
        'conn_fc': {},
        'conn_ff': {},
        'connection_type': 'none'
    }


def load_topology(infile):
    with open(infile) as LD:
        topology = json.load(LD)
    return topology


def save_topology(topology, outfile):
    with open(outfile, "w") as fout:
        json.dump(topology, fout, indent=4)


def set_connection_type(topology, conntype):
    topology['connection_type'] = conntype


def get_connection_label(src, dst):
    return '%s_%s' % (src, dst)


def terminals_from_connection_label(lbl):
    return lbl.split('_')


def delaytype_from_nodetype(nodetype_src, nodetype_dst):
    delaytype = None
    if nodetype_src == 'sensors' and nodetype_dst == 'fog':
        delaytype = 'delay_sf'
    if nodetype_src == 'fog' and nodetype_dst == 'cloud':
        delaytype = 'delay_fc'
    if nodetype_src == 'fog' and nodetype_dst == 'fog':
        delaytype = 'delay_ff'
    return delaytype


def conntype_from_nodetype(nodetype_src, nodetype_dst):
    conntype = None
    if nodetype_src == 'sensors' and nodetype_dst == 'fog':
        conntype = 'delay_sf'
    if nodetype_src == 'fog' and nodetype_dst == 'cloud':
        conntype = 'delay_fc'
    if nodetype_src == 'fog' and nodetype_dst == 'fog':
        conntype = 'delay_ff'
    return conntype


def delaytype_from_conntype(conntype):
    delaytype = None
    if conntype == 'conn_sf':
        delaytype = 'delay_sf'
    if conntype == 'conn_fc':
        delaytype = 'delay_fc'
    if conntype == 'conn_ff':
        delaytype = 'delay_ff'
    return delaytype


def convert_topology(topology):
    new_topology = create_topology()

    for i in topology["Cloud"]:
        new_topology["cloud"][i["ID"]] = {"lat": i["Latitudine"],
                                          "lon": i["Longitudine"]}

    for i in topology["Nodi Fog"]:
        new_topology["fog"][i["ID"]] = {"lat": i["Latitudine"],
                                        "lon": i["Longitudine"]}

    for i in topology["Sensori"]:
        new_topology["sensors"][i["ID"]] = {"lat": i["Latitudine"],
                                            "lon": i["Longitudine"]}
    compute_distances(new_topology, 'sensors', 'fog')
    compute_distances(new_topology, 'fog', 'cloud')
    compute_distances(new_topology, 'fog', 'fog')

    return new_topology


######################################################################################################################
# Management of nodes
######################################################################################################################
def get_nodes_keys(topology, nodetype):
    if topology[nodetype] is not None:
        return topology[nodetype].keys()


def set_node(topology, nodetype, i, lat, lon, address="unknown"):
    topology[nodetype][i] = {'lat': lat, 'lon': lon, 'address': address}


def get_node_latitude(topology, nodetype, i):
    return float(topology[nodetype][i]['lat'])


def get_node_longitude(topology, nodetype, i):
    return float(topology[nodetype][i]['lon'])


def get_node_attr(topology, nodetype, i, attrname):
    return topology[nodetype][i][attrname]


def set_node_attr(topology, nodetype, i, attrname, val):
    topology[nodetype][i][attrname] = val


######################################################################################################################
# Management of distances and delays
######################################################################################################################
def calc_dist(a, b, c, d):
    lon1 = float(a)
    lat1 = float(b)
    lon2 = float(c)
    lat2 = float(d)
    # radians -> deg
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    r = 6371  # earth radius (KM)
    return c * r


def compute_distances(topology, nodetype_src, nodetype_dst):
    for s in get_nodes_keys(topology, nodetype_src):
        for d in get_nodes_keys(topology, nodetype_dst):
            set_delay(topology, nodetype_dst=nodetype_dst, nodetype_src=nodetype_src, src=s, dst=d,
                      delay=calc_dist(get_node_latitude(topology, nodetype_src, s),
                                      get_node_longitude(topology, nodetype_src, s),
                                      get_node_latitude(topology, nodetype_dst, d),
                                      get_node_longitude(topology, nodetype_dst, d)))


def get_avg_delay(topology, distancetypes_list):
    dly = 0.0
    n = 0
    for d in distancetypes_list:
        for i in get_delays(topology, d):
            dly += get_delay(topology, delaytype=d, lbl=i)
            n += 1
    return dly / n


def scale_delay(topology, delaytypes_list, k):
    for d in delaytypes_list:
        for i in get_delays(topology, d):
            set_delay(topology, delaytype=d, lbl=i, delay=get_delay(topology, delaytype=d, lbl=i) * k)


def normalize_delay(topology, delaytypes_list, delta):
    d = get_avg_delay(topology, delaytypes_list)
    scale_delay(topology, delaytypes_list, delta / d)


def get_delays(topology, delaytype, src='*', dst='*'):
    if topology[delaytype] is None:
        return None
    ks = topology[delaytype].keys()
    if src == '*' and dst == '*':
        return ks
    rv = []
    for i in ks:
        if src == '*' or src == get_delay_source(topology, delaytype=delaytype, lbl=i):
            rv.append(i)
    return rv


def get_delay(topology, delaytype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if delaytype is None:
        delaytype = delaytype_from_nodetype(nodetype_src, nodetype_dst)
    return float(topology[delaytype][lbl]['delay'])


def get_delay_source(topology, delaytype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if delaytype is None:
        delaytype = delaytype_from_nodetype(nodetype_src, nodetype_dst)
    return str(topology[delaytype][lbl]['src'])


def get_delay_destination(topology, delaytype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if delaytype is None:
        delaytype = delaytype_from_nodetype(nodetype_src, nodetype_dst)
    return str(topology[delaytype][lbl]['dst'])


def set_delay(topology, delaytype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None, delay=0.0):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if delaytype is None:
        delaytype = delaytype_from_nodetype(nodetype_src, nodetype_dst)
    if (src is None) or (dst is None):
        (src, dst) = terminals_from_connection_label(lbl)
    topology[delaytype][lbl] = {'delay': delay, 'src': src, 'dst': dst}


######################################################################################################################
# Management of connections
######################################################################################################################
def get_connection_keys(topology, conntype):
    if topology[conntype] is not None:
        return topology[conntype].keys()


def get_connection_delay(topology, conntype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if conntype is None:
        conntype = conntype_from_nodetype(nodetype_src, nodetype_dst)
    if (conntype not in topology) or (lbl not in topology[conntype]):
        return None
    else:
        return float(topology[conntype][lbl]['delay'])


def get_connection_source(topology, conntype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if conntype is None:
        conntype = conntype_from_nodetype(nodetype_src, nodetype_dst)
    if (conntype not in topology) or (lbl not in topology[conntype]):
        return None
    else:
        return topology[conntype][lbl]['src']


def get_connection_destination(topology, conntype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None,
                               dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if conntype is None:
        conntype = conntype_from_nodetype(nodetype_src, nodetype_dst)
    if (conntype not in topology) or (lbl not in topology[conntype]):
        return None
    else:
        return topology[conntype][lbl]['dst']


def set_connection(topology, conntype=None, nodetype_src=None, nodetype_dst=None, lbl=None, src=None, dst=None):
    if lbl is None:
        lbl = get_connection_label(src, dst)
    if (src is None) or (dst is None):
        (src, dst) = terminals_from_connection_label(lbl)
    lbldelay = get_connection_label(src, dst)
    if conntype is None:
        conntype = conntype_from_nodetype(nodetype_src, nodetype_dst)
    if conntype not in topology:
        topology[conntype] = {}
    delaytype = delaytype_from_conntype(conntype)
    d = get_delay(topology, delaytype=delaytype, lbl=lbldelay)
    topology[conntype][lbl] = {'delay': d, 'src': src, 'dst': dst}
